open Coq_checker
open Ast

(* conversion des types Caml <-> Types Coq *)
let list_to_coq_list l = 
  let rec l2cl acc = function
    | [] -> acc
    | a::l -> l2cl (Cons (a, acc)) l
  in
    l2cl Nil (List.rev l)

let coq_list_to_list l = 
  let rec cl2l acc = function
    | Nil -> acc
    | Cons(a, l) -> cl2l (a::acc) l
  in
    List.rev (cl2l [] l)

let prod_to_coq_prod (x, y) = Pair (x, y)
  
let coq_prod_to_prod = function Pair (x, y) -> x, y

let rec positive_to_int p =
  match p with
    | XH -> 1
    | XO p' -> 2 * (positive_to_int p')
    | XI p' -> 2 * (positive_to_int p') + 1

let rec int_to_positive i =
  if (i <= 0) then
    failwith ((string_of_int i)^" cannot be converted into positive")
  else if (i = 1) then
    XH
  else (
      if ((i land 1) = 1) then
	XI (int_to_positive (i lsr 1))
      else
	XO (int_to_positive (i lsr 1))
    )
    
let positive_to_string p = string_of_int (positive_to_int p)

(* Fonction qui update un environnement cod par t : 'a list ref 
   Elle ajoute [f] dans la liste [t] si il n'y est pas, puis renvoie la
   position de [f] dans [t].
*)
let symb_to_int t f = 
  let rec pos f n = function
    | [] -> ([f], n)
    | a::l ->
	let (l', n') = 
	  if (a = f) then (l, n) else pos f (n+1) l
	in
	  a::l', n'
  in
  let (l, n) = pos f 1 (!t) in
    t := l; n

let int_to_symb t i =
  let rec get = function
    | _, [] | 0, _ -> failwith ((string_of_int i)^"has no symbol")
    | j, a::l -> if (j = 1) then a else get (j-1, l)
  in get (i, !t)



(* table des symboles de fonctions : 'a -> int *)
let f_symb = ref []
let f_symb_to_pos f = int_to_positive (symb_to_int f_symb f)
let pos_to_f_symb p = int_to_symb f_symb (positive_to_int p)

(* table des variables : 'a -> int *)
let v_symb = ref []
let v_symb_to_pos v = int_to_positive (symb_to_int v_symb v)
let pos_to_v_symb p = int_to_symb v_symb (positive_to_int p)

(* table des tats : 'a -> int *)
let s_symb = ref []
let state_to_pos e = int_to_positive (symb_to_int s_symb e)
let pos_to_state p = int_to_symb s_symb (positive_to_int p)

let rec to_cterm var = function
  | Ast.T (id, tl) when (var id.raw)->
      Coq_checker.Var (v_symb_to_pos id.raw)

  | Ast.T (id, tl) ->
      Coq_checker.Fun
	( f_symb_to_pos id.raw,
	  (list_to_coq_list (List.map (to_cterm var) tl))
	)
  
  | VarSpe id -> Coq_checker.Var (v_symb_to_pos id.raw)


(* Guarded : constraints not considered for now *)
let to_crule var = function
  | Simple (lhs, rhs) | Guarded (lhs, rhs, _) ->
      Lrule (to_cterm var lhs, to_cterm var rhs)

let norm_transition = function
  | Simple  (T(id, tl), T (st, _))
  | Guarded (T(id, tl), T (st, _), _) ->
      ( f_symb_to_pos id.raw,
	list_to_coq_list
	  (List.map
	     ( function 
		 | Ast.T (id, _) -> state_to_pos id.raw
		 | _ -> failwith "Must not happen : Bad transition"
	     )
	     tl
	  ),
	state_to_pos st.raw
      )
  | _ -> failwith "Must not happen : Bad Transition"


let extract_f_states aut =
  list_to_coq_list
    ( List.map
	(fun x -> state_to_pos x.raw)
	aut.f_states
    )
    
let build_delta (aut : Ast.automaton) = 
  let rec build_delta (d: Delta.t) = function
    | [] -> d
    | ext::l ->
	let f, ls, q = norm_transition ext in
	  build_delta (Delta.add q f ls d) l
  in
    build_delta
      (build_delta Delta.empty aut.prior)
      aut.rules

let build_automaton a =
  { qf = extract_f_states a;
    d = build_delta a
  }


let build_trs var (_, trs : Ast.trs) =
  (* Pour l'instant on ne traite pas les conditions dans trs *)
  list_to_coq_list (List.map (to_crule var) trs)


(* Code for debug *)
let delta_to_string (d : Delta.t) =
  Delta.fold
    (fun str q f cl ->
       let q = pos_to_state q
       and f = pos_to_f_symb f
       and cl = 
	 (List.fold_left
	    (fun s q -> s^(pos_to_state q)^" ,")
	    "( "
	    (coq_list_to_list cl)
	 )^"\b)"
       in
	 str^f^cl^" -> "^q^"\n"
    ) d ""

let dump_delta d = 
  Format.printf "%s\n" (delta_to_string d);
  Format.printf "%i transitions\n"
    (Delta.fold (fun i _ _ _ -> i + 1) d 0);
  Format.printf "%i states"
    (List.length (coq_list_to_list (Delta.dom d)))
(* Exemple *)

let rec context_to_string = function
  | Var x -> pos_to_v_symb x
  | Fun (f, lc) ->
      let l' = List.map context_to_string (coq_list_to_list lc) in
	(List.fold_left (fun s c -> s^c^" ,") ((pos_to_f_symb f)^"(") l')^"\b)"
  | _ -> failwith "Must not happen  : no state in context"

let fstates_to_string l = 
  let l' = List.map pos_to_state (coq_list_to_list l) in
    List.fold_left (fun s p -> s^" "^p) "" l'

let dump_fstates fstates = 
  Format.printf "%s\n" (fstates_to_string fstates)

let rule_to_string l r =
  "[ "^(context_to_string l)^" -> "^
	(context_to_string r)^" ]"

let trs_to_string trs =
  List.fold_left
    (fun s (Lrule(l, r)) -> 
       s^"[ "^(context_to_string l)^" -> "^
	 (context_to_string r)^" ]\n")
    ""
    (coq_list_to_list trs)

let trs_count trs =
  List.length 
    (coq_list_to_list trs)


let dump_trs trs =
  Format.printf "%s\n" (trs_to_string trs);
  Format.printf "%i rules found" (trs_count trs)
  
(* Petits essais
let (_, trs), (_, a, fsa), (_, b, fsb) = load_for_check "ex1_comp.txt";;

let (Pair (l, r))::_ = (coq_list_to_list trs);;

let t1 = Checker_coq.check_rule_at_q a (int_to_positive 3) l r
let res = Checker_coq.check_rules b trs

let (_, trs), (_, a, fsa), (_, b, fsb) = load_for_check "nspk.txt";;

*)
